/*
 * This is Tampa Ping-Pong convers/conversd derived from the wampes
 * convers package written by Dieter Deyke <deyke@mdddhd.fc.hp.com>
 * Modifications by Fred Baumgarten <dc6iq@insu1.etec.uni-karlsruhe.de>
 *
 * Modifications by Brian A. Lantz/KO4KS <brian@lantz.com> 
 */

#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <utmp.h>
#include <syslog.h>
#include <pwd.h>

#include "conversd.h"
#include "user.h"
#include "host.h"
#include "tnos.h"
#include "convert.h"
#include "version.h"
#include "log.h"
#include "cfgfile.h"
#include "config.h"
#include "access.h"
#include "hmalloc.h"

fd_set chkread;
fd_set chkwrite;

char convtype[512];
time_t boottime;
time_t currtime;
char userfile[128];
char motdfile[128];
char restrictfile[128];
char accessfile[128];
char noaccessfile[128];
char helpfile[128];
char observerfile[128];
char issuefile[128];
char pidfile[128];
int NR_PERMLINKS = DEF_NR_PERMLINKS;
char myfeatures[] = "Aadmpun";
char OBSERVERSTR[] = "*** Sorry, you are only an observer :-)\n";
char OBSERVERSTR2[] = "*** That command is not available to observers....\n";
char OBSERVERSTR3[] = "*** Observers are not permitted to send messages....\n";
char OBSERVERID[] = "Observer only";

static int maxfd = -1;

struct connection *connections;
struct permlink **permarray = NULLPERMARRAY;
struct destination *destinations;
struct channel *channels;

struct extendedcmds *ecmds;

struct stats daily[60];
struct stats hourly[60];
time_t starttime;
time_t nextdailystats;
time_t nexthourlystats;
long hoursonline = 0L;
long daysonline = 0L;

static void process_input(struct connection * cp);

extern void rip(char *str);

struct cmdtable cmdtable[] =
{

	{"/quit",	bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
	{"?",		cmdsummary_command,	CM_USER | CM_OBSERVER},
	{"away",	away_command,		CM_USER},
	{"action",	me_command,		CM_USER},
	{"all",		all_command,		CM_USER},
	{"beep",	beep_command,		CM_USER | CM_OBSERVER},
	{"bell",	beep_command,		CM_USER | CM_OBSERVER},
	{"bjack",       bjack_command,          CM_USER},
	{"bye",		bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
	{"channel",	channel_command,	CM_USER | CM_OBSERVER},
	{"charset",	charset_command,	CM_USER | CM_OBSERVER},
/*	{"cstat",	cstat_command,		CM_UNKNOWN | CM_USER | CM_OBSERVER}, */
	{"cut",         cut_command,            CM_USER | CM_OBSERVER},
	{"destinations", hosts_command,		CM_USER | CM_OBSERVER},	
	{"exit",	bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
	{"exclude",	imsg_command,		CM_USER},
	{"help",	help_command,		CM_USER | CM_OBSERVER},
	{"halt",	restart_command,	CM_USER},
	{"hosts",	hosts_command,		CM_USER | CM_OBSERVER},
	{"invite",	invite_command,		CM_USER},
	{"imsg",	imsg_command,		CM_USER},
	{"join",	channel_command,	CM_USER | CM_OBSERVER},
	{"links",	links_command,		CM_USER | CM_OBSERVER},
	{"last",	last_command,		CM_USER | CM_OBSERVER},
	{"leave",	leave_command,		CM_USER | CM_OBSERVER},
	{"list",	list_command,		CM_USER | CM_OBSERVER},
	{"msg",		msg_command,		CM_USER},
	{"me",		me_command,		CM_USER},
	{"mode",	mode_command,		CM_USER},
	{"name",	name_command,		CM_UNKNOWN},
	{"news",	news_command,		CM_USER | CM_OBSERVER},
	{"nickname",	nickname_command,	CM_USER},
	{"nonickname",	nonickname_command,	CM_USER},
	{"notify",	notify_command,		CM_USER},
	{"observer",	observer_command,	CM_UNKNOWN},
	{"operator",	operator_command,	CM_USER},
	{"online",	who_command2,		CM_UNKNOWN},
	{"personal",	personal_command,	CM_USER},
	{"prompt",	prompt_command,		CM_USER},
	{"quit",	bye_command,		CM_USER | CM_OBSERVER | CM_UNKNOWN},
	{"query",	query_command,		CM_USER},
/*	{"quote",	quote_command,		CM_USER | CM_OBSERVER},*/
	{"realname",	realname_command,	CM_USER | CM_OBSERVER},
	{"restart",	restart_command,	CM_USER},
	{"restricted",	restricted_command,	CM_UNKNOWN},
	{"roll",        roll_command,           CM_USER | CM_OBSERVER},
	{"smiley",	smiley_command,		CM_USER},
	{"spam",	spam_command,		CM_USER},
	{"stats",	stats_command,		CM_USER | CM_OBSERVER},
	{"sysinfo",	sysinfo_command,	CM_USER | CM_OBSERVER},
	{"sysop",	operator_command,	CM_USER},
	{"time",	time_command,		CM_USER},
	{"topic",	topic_command,		CM_USER},
	{"uptime",	uptime_command,		CM_USER | CM_OBSERVER},
	{"verbose",	verbose_command,	CM_USER},
	{"version",	version_command,	CM_USER | CM_OBSERVER},
	{"who",		who_command,		CM_USER | CM_OBSERVER},
	{"whois",	whois_command,		CM_USER | CM_OBSERVER},
	{"width",	width_command,		CM_USER | CM_OBSERVER},
	{"wx",		wx_command,		CM_USER | CM_OBSERVER},

	{"\377\200away", h_away_command,	CM_HOST},
	{"\377\200cmsg", h_cmsg_command,	CM_HOST},
	{"\377\200dest", h_dest_command,	CM_HOST},
	{"\377\200ecmd", h_ecmd_command,	CM_HOST},
	{"\377\200help", h_help_command,	CM_HOST},
	{"\377\200host", h_host_command,	CM_UNKNOWN},
	{"\377\200invi", h_invi_command,	CM_HOST},
	{"\377\200link", h_link_command,	CM_HOST},
	{"\377\200loop", h_loop_command,	CM_HOST},
	{"\377\200mode", mode_command,		CM_HOST},
	{"\377\200obsv", h_user_command,	CM_HOST},
	{"\377\200oper", h_oper_command,	CM_HOST},
	{"\377\200ping", h_ping_command,	CM_HOST},
	{"\377\200pong", h_pong_command,	CM_HOST},
	{"\377\200rout", h_rout_command,	CM_HOST},
	{"\377\200sysi", h_sysi_command,	CM_HOST},
	{"\377\200topi", h_topi_command,	CM_HOST},
	{"\377\200uadd", h_uadd_command,	CM_HOST},
	{"\377\200udat", h_udat_command,	CM_HOST},
	{"\377\200umsg", h_umsg_command,	CM_HOST},
	{"\377\200user", h_user_command,	CM_HOST},

	{NULL, 0, 0}
};

static char *getargp;

void setargp(char *str)
{
	getargp = str;
}

/* -----------------------------------------------------------------------
 *	Log a link command
 */

void linklog(struct connection *p, int direction, const char *fmt, ...)
{
#if 0
	va_list args;
	char s[LOG_LEN];
	
	va_start(args, fmt);
#ifdef NO_SNPRINTF
	vsprintf(s, fmt, args);
#else
	vsnprintf(s, LOG_LEN, fmt, args);
#endif
	va_end(args);
	s[LOG_LEN-1] = '\0';

	/* TODO: write to per-link file */
	fprintf(stderr, "%s %-10.10s %s", (direction == L_SENT) ? "=>" : "<=", p->name, s);
#endif
}

/* -----------------------------------------------------------------------
 *	Write a string to a socket, with character conversion (for users)
 *	and extra buffering
 */

void appendstring(struct connection *cp, const char *string)
{
	struct mbuf *bp, *p;
	char buffer[4096];
	const char *p_string;
	
	if (!*string)
		return;
	
	if ((cp->type == CT_USER) && (strncmp(string, "/\377\200", 3))) {
		strncpy(buffer, string, 4096);
		convert(ISO, cp->charset_out, buffer);
		p_string = buffer;
	} else
		p_string = string;
	
	bp = (struct mbuf *) hmalloc(sizeof(*bp) + strlen(p_string) + 1);
	bp->next = 0;
	bp->data = strcpy((char *) (bp + 1), p_string);
	
	if (cp->obuf) {
		for (p = cp->obuf; p->next; p = p->next);
		p->next = bp;
	} else {
		cp->obuf = bp;
		FD_SET(cp->fd, &chkwrite);
	}
	
	if (queuelength(cp->obuf) > ((cp->type == CT_USER) ? max_u_queue : max_h_queue))
		bye_command(cp);
}

/*---------------------------------------------------------------------------*/

void appendc(struct connection *cp, const int n, const int ast)
{
	static char x[2];
	char c;
	
	c = cp->prompt[n];
	if (c == '\0') {
		if (ast) {
			appendstring(cp, "***\n");
		}
	} else {
		x[0] = c;
		x[1] = '\0';
		appendstring(cp, x);
	}
}


/* -----------------------------------------------------------------------
 *	
 */

void appendprompt(struct connection *cp, const int ast)
{
	if (*cp->query)
		appendc(cp, 1, ast);
	else
		appendc(cp, 2, ast);
		
	fflush(NULL);
}

/* -----------------------------------------------------------------------
 *	Calculate total queue length
 */

int queuelength(const struct mbuf *bp)
{
	int len;
	
	for (len = 0; bp; bp = bp->next)
		len += strlen(bp->data);
		
	return len;
}

/* -----------------------------------------------------------------------
 *	Find a permlink, given it's name, or NULL if no named permlink
 *	wkt@cs.adfa.oz.au
 */
 
struct permlink *find_permlink(const char *name)
{
	int pl;
	struct permlink *p;
	
	if (permarray)
		for (pl = 0; pl < NR_PERMLINKS; pl++) {
			p = permarray[pl];
			if (p && !strcasecmp(p->name, name))
				return (p);
		}
		
	return (NULL);
}

/* -------------------------------------------------------------------------
 *	Free a channel structure
 */

void destroy_channel(int number)
{
	struct channel *ch, *ch1;

	ch1 = NULLCHANNEL;
	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == number) {
			if (ch1) {
				ch1->next = ch->next;
			} else {
				channels = ch->next;
			}
			hfree(ch);
			break;
		}
		ch1 = ch;
	}
}

/* ------------------------------------------------------------------------
 *	Check that the destination is not me, is defined, has a link,
 *	and via this connection.
 */
 
int dest_check(struct connection *cp, struct destination *dest)
{
	int retval = 0;
	if (dest == NULLDESTINATION || dest->link == NULLPERMLINK)
		retval = 1;
	else {
		if (strcasecmp(dest->name, myhostname) && dest && dest->link)
			retval = ((dest->link->connection == cp) ? 1 : 0);
	}
	
	return retval;
}


/* ------------------------------------------------------------------------
 *	Check for loop on the user
 */

int user_check(struct connection *cp, char *user)
{
	struct connection *p;
	for (p = connections; p; p = p->next) {
		if ((p->type == CT_USER) && (!strcasecmp(p->name, user) || !strcasecmp(p->nickname, user)))
			return (p->via && cp != p->via) ? 0 : 1;
	}
	return 1;
}

/* -----------------------------------------------------------------------
 *	Reset destination routes which are not connected, do not have
 *	a link defined, or are from this connection.
 */

void route_cleanup(struct connection *cp)
{
	static struct destination *d;
	
	for (d = destinations; d; d = d->next) {
		if ((!d->link) || (!d->connected) || (!d->link->connection) || (!d->link->connection->name) || (!cp->name)
		 || (!strcasecmp(d->link->connection->name, cp->name))) {
			d->link = NULLPERMLINK;
			d->rtt = 0;
			d->last_sent_rtt = 0;
			d->connected = 0;
		}
	}
}

/*---------------------------------------------------------------------------*/

int ecmd_exists(char *cmdname)
{
	struct extendedcmds *echk;

	echk = ecmds;
	while (echk && echk->name) {
		if (!strcasecmp(echk->name, cmdname))
			return 1;
		echk = echk->next;
	}
	return 0;
}

/*---------------------------------------------------------------------------*/

void free_connection(struct connection *cp)
{
	struct mbuf *bp;
	struct permlink *p = NULLPERMLINK;
	int pl = -1;

	if (permarray)
		for (pl = 0; pl < NR_PERMLINKS; pl++) {
			p = permarray[pl];
			if (p && (p->connection == cp)) {
				p->connection = NULLCONNECTION;
				break;
			}
		}
		
	if (cp->fd >= 0) {
		close(cp->fd);
		FD_CLR(cp->fd, &chkread);
		FD_CLR(cp->fd, &chkwrite);
		if (cp->fd == maxfd)
			while (--maxfd >= 0)
				if (FD_ISSET(maxfd, &chkread))
					break;
		cp->fd = -1;
	}
	while ((bp = cp->obuf) != NULL) {
		cp->obuf = bp->next;
		hfree(bp);
	}
	hfree(cp);
	if ((pl > -1) && (pl < NR_PERMLINKS)) {
		if (p && !p->permanent) {
			permarray[pl] = NULLPERMLINK;
		}
	}
}

/*---------------------------------------------------------------------------*/

void free_closed_connections()
{
	struct connection *cp, *p;
	struct permlink *l;
	int pl;

	for (p = NULLCONNECTION, cp = connections; cp;) {
		l = NULLPERMLINK;
		if (permarray)
			for (pl = 0; pl < NR_PERMLINKS; pl++) {
				if ((permarray[pl]) && (permarray[pl]->connection == cp)) {
					l = permarray[pl];
					break;
				}
			}
			
		if (l) {
			if (cp->type == CT_CLOSED ||
			    (cp->type == CT_UNKNOWN && cp->time + l->waittime - 5 < currtime)) {
				if (p) {
					p->next = cp->next;
					free_connection(cp);
					cp = p->next;
				} else {
					connections = cp->next;
					free_connection(cp);
					cp = connections;
				}
			} else {
				p = cp;
				cp = cp->next;
			}
		} else {
			if (cp->type == CT_CLOSED) {
				if (p) {
					p->next = cp->next;
					free_connection(cp);
					cp = p->next;
				} else {
					connections = cp->next;
					free_connection(cp);
					cp = connections;
				}
			} else {
				p = cp;
				cp = cp->next;
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

char *getarg(char *line, int all)
{
	char *arg;
	int c;

	if (line)
		getargp = line;
	while (isspace(uchar(*getargp)))
		getargp++;
	if (all)
		return getargp;
	arg = getargp;
	while (*getargp && !isspace(uchar(*getargp))) {
		c = tolower(uchar(*getargp));
		*getargp++ = c;
	}
	if (*getargp)
		*getargp++ = '\0';
	return arg;
}

/*---------------------------------------------------------------------------*/

char *getargcs(char *line, int all)
{
	char *arg;

	if (line)
		getargp = line;
	while (isspace(uchar(*getargp)))
		getargp++;
	if (all)
		return getargp;
	arg = getargp;
	while (*getargp && !isspace(uchar(*getargp))) {
		getargp++;
	}
	if (*getargp)
		*getargp++ = '\0';
	return arg;
}

/*---------------------------------------------------------------------------*/

void getTXname(struct connection *cp, char *buf)
{
	if (!strcasecmp(cp->nickname, cp->name))
		strcpy(buf, cp->name);
	else
		sprintf(buf, "%s:%s", cp->name, cp->nickname);
}


/*---------------------------------------------------------------------------*/

char *formatline(char *prefix, char *text, int linelen)
{
	char *f, *t, *x;
	int prefixlen, l, lw;
	static char buf[2048];

	linelen--;
	prefixlen = strlen(prefix);
	for (f = prefix, t = buf; *f; *t++ = *f++);
	l = t - buf;
	f = text;

	for (;;) {
		while (isspace(uchar(*f)))
			f++;
		if (!*f) {
			*t++ = '\n';
			*t = '\0';
			return buf;
		}
		for (x = f; *x && !isspace(uchar(*x)); x++);
		lw = x - f;
		if (l > prefixlen && l + 1 + lw > linelen) {
			*t++ = '\n';
			l = 0;
		}
		do {
			*t++ = ' ';
			l++;
		} while (l < prefixlen);
		while (lw--) {
			*t++ = *f++;
			l++;
		}
	}
}

/*---------------------------------------------------------------------------*/

char *ts(time_t gmt)
{
	static char buffer[80];
	static char monthnames[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
	struct tm *tm;

	tm = localtime(&gmt);
	if (gmt + 24 * 60 * 60 > currtime)
		sprintf(buffer, " %2d:%02d", tm->tm_hour, tm->tm_min);
	else
		sprintf(buffer, "%-3.3s %2d", monthnames + 3 * tm->tm_mon, tm->tm_mday);
	return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts2(time_t gmt)
{
	static char buffer[80];
	static char monthnames[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
	struct tm *tm;

	tm = localtime(&gmt);
	if (tm) {
		if (gmt + 24 * 60 * 60 > currtime)
			sprintf(buffer, "%2d:%02d", tm->tm_hour, tm->tm_min);
		else
			sprintf(buffer, "%-3.3s %2d", monthnames + 3 * tm->tm_mon, tm->tm_mday);
	} else
		strcpy(buffer, "?????");
	return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts3(time_t seconds, char *buffer)
{
	if (seconds < 100) {
		sprintf(buffer, "%lds", (long)seconds);
	} else if (seconds < 6000) {
		sprintf(buffer, "%ldm", (long)seconds / 60);
	} else if (seconds < 360000) {
		sprintf(buffer, "%ldh", (long)seconds / 3600);
	} else {
		sprintf(buffer, "%ldd", (long)seconds / 86400);
	}
	return buffer;
}

/*---------------------------------------------------------------------------*/

char *ts4(time_t seconds)
{
	int days, hours, minutes;
	static char buffer[2048];
	char buf[128];

	days = seconds / 86400;
	seconds -= days * 86400;
	hours = seconds / 3600;
	seconds -= hours * 3600;
	minutes = seconds / 60;
	seconds -= minutes * 60;
	buffer[0] = '\0';
	if (days) {
		sprintf(buf, "%d day%s, ", days, (days == 1) ? "" : "s");
		strncpy(buffer, buf, 2048);
	}
	if (days + hours) {
		sprintf(buf, "%d hour%s, ", hours, (hours == 1) ? "" : "s");
		strcat(buffer, buf);
	}
	if (days + hours + minutes) {
		sprintf(buf, "%d minute%s, ", minutes, (minutes == 1) ? "" : "s");
		strcat(buffer, buf);
	}
	sprintf(buf, "%ld second%s.", (long)seconds, (seconds == 1) ? "" : "s");
	strcat(buffer, buf);
	return buffer;
}

/*---------------------------------------------------------------------------*/

struct connection *alloc_connection(int fd)
{
	struct connection *cp;
	
	cp = (struct connection *)hcalloc(1, sizeof(struct connection));
	cp->fd = fd;
	cp->time = currtime;
	cp->verbose = 0;
	cp->prompt[0] = 0;
	cp->prompt[1] = 0;
	cp->prompt[2] = 0;
	cp->prompt[3] = 0;
	cp->features = 0;
	*cp->rev = '\0';
	*cp->away = '\0';
	cp->chan_list = NULLCLIST;
	cp->amprnet = 1;
	cp->oldaway = 1;
	cp->atime = currtime;
	cp->mtime = 0;
	*cp->pers = '\0';
	cp->width = 80;
	cp->observer = 0;
	cp->hostallowed = 1;
	cp->next = connections;
	connections = cp;
	if (maxfd < fd)
		maxfd = fd;
	FD_SET(fd, &chkread);
	return cp;
}

/*---------------------------------------------------------------------------*/

void accept_connect_request(int flisten)
{
	FILE *fdi;
	char buffer[2048];
	int addrlen;
	int fd, flags;
	struct sockaddr addr;
	struct connection *cp;
	struct sockaddr_in *sin = (struct sockaddr_in *) &addr;
	int allow = 0;
	
	addrlen = sizeof(addr);
	if ((fd = accept(flisten, &addr, &addrlen)) < 0)
		return;
	if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
		log(L_ERR, "accept_connect_request(): fcntl(F_GETFL) failed: %s", strerror(errno));
		close(fd);
		return;
	}
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
		log(L_ERR, "accept_connect_request(): fcntl(F_SETFL, O_NONBLOCK) failed: %s", strerror(errno));
		close(fd);
		return;
	}
	if (sin->sin_family == AF_INET) {
		log(L_SOCKET, "Incoming connection from [%s]", inet_ntoa(sin->sin_addr));
		allow = chk_access(sin->sin_addr.s_addr);
		if (!allow) {
			log(L_AUTH, "Refused connection from [%s]", inet_ntoa(sin->sin_addr));
			close(fd);
			return;
		}
	} else {
		log(L_ERR, "Gee, connection with non AF_INET socket! Are we not in Kansas any more?");
		close(fd);
		return;
	}
	
	cp = alloc_connection(fd);
	memcpy(&cp->sin, sin, sizeof(struct sockaddr_in));
	
	if (!(allow & ACC_HOST))
		cp->hostallowed = 0;
	if (allow & ACC_OBSRV)
		cp->observer = 1;
	if (allow & ACC_RESTR)
		cp->type = -1;
	
	if (!access(issuefile, R_OK)) {
		fdi = fopen(issuefile, "r");
		buffer[0] = '*';
		buffer[1] = ' ';
		while (!feof(fdi)) {
			fgets(buffer + 2, 2045, fdi);
			if (!feof(fdi)) {
				appendstring(cp, buffer);
			}
		}
		fclose(fdi);
	}
}

/*---------------------------------------------------------------------------*/

int count_user(int channel)
{
	struct connection *p;
	struct clist *cl;
	int n = 0;

	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER) {
			if (p->via && (channel == -1 || p->channel == channel)) {
				n++;
			} else {
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (channel == -1 || cl->channel == channel) {
						n++;
						break;
					}
				}
			}
		}
	}
	return n;
}

/*---------------------------------------------------------------------------*/

int count_topics(void)
{
	int n = 0;
	struct channel *ch;

	for (ch = channels; ch; ch = ch->next)
		if (!(ch->flags & (M_CHAN_S || M_CHAN_I)) && *ch->topic)
			n++;
	return n;
}

/*---------------------------------------------------------------------------*/

int isonchannel(struct connection *cp, char *user, char *host)
{
	struct connection *p;
	struct clist *cl, *cl1;

	for (p = connections; p; p = p->next) {
		if ((p->type == CT_USER) && (!p->via) && (!strcasecmp(p->name, user))) {
			for (cl1 = cp->chan_list; cl1; cl1 = cl1->next) {
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cl->channel == cl1->channel)
						return 1;
				}
			}
		}
	}
	return 0;
}

/*---------------------------------------------------------------------------*/

void clear_locks()
{
	struct connection *p;

	for (p = connections; p; p = p->next)
		p->locked = 0;
}

/*---------------------------------------------------------------------------*/

void notify_destinations(struct destination *dest, 
	const char *host, const char *rev, int state)
{
	struct connection *p;
	char buffer[2048];
	char revstr[40];

	if (dest && ((state && dest->connected) || (!state && !dest->connected)))
		return;
	clear_locks();
	if (rev && *rev)
		sprintf(revstr, "(%.10s)", rev);
	else
		revstr[0] = 0;
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER && !p->via && p->operator == 2) {
			sprintf(buffer, "*** (%s) Server %sing the network: %s %s\n", ts2(currtime), (state) ? "join" : "leav", host, revstr);
			appendstring(p, buffer);
			appendprompt(p, 0);
		}
	}
	if (dest) {
		dest->connected = state;
		if (state == 0) {
			dest->link = NULLPERMLINK;
			dest->rtt = 0;
			dest->last_sent_rtt = 0;
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_awaymsg(char *fromname, char *hostname, time_t time, char *text)
{
	char buffer[2048];
	struct connection *p;

	for (p = connections; p; p = p->next) {
		if (p->type == CT_HOST) {
			if (!p->locked) {
				if (p->oldaway) {
					sprintf(buffer, "/\377\200AWAY %s %s 69 %ld %s\n", fromname, hostname, (long)time, text);
				} else {
					sprintf(buffer, "/\377\200AWAY %s %s %ld %s\n", fromname, hostname, (long)time, text);
				}
				linklog(p, L_SENT, "%s", &buffer[3]);
				appendstring(p, buffer);
				p->locked = 1;
			}
		} else {
			if (!p->locked && !p->via) {
				if (isonchannel(p, fromname, hostname)) {
					appendc(p, 0, 0);
					appendc(p, 3, 0);
					if (*text != '\0')
						sprintf(buffer, "*** (%s) %s@%s has gone away:\n    %s\n", ts2(currtime), fromname, hostname, text);
					else
						sprintf(buffer, "*** (%s) %s@%s is back again.\n", ts2(currtime), fromname, hostname);
					appendstring(p, buffer);
					appendprompt(p, 0);
					p->locked = 1;
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_mode(struct channel *ch)
{
	struct connection *p;
	char buffer[2048];
	char flags[16];

	flags[0] = '\0';
	if (ch->flags & M_CHAN_S) {
		strcat(flags, "S");
	}
	if (ch->flags & M_CHAN_P) {
		strcat(flags, "P");
	}
	if (ch->flags & M_CHAN_T) {
		strcat(flags, "T");
	}
	if (ch->flags & M_CHAN_I) {
		strcat(flags, "I");
	}
	if (ch->flags & M_CHAN_M) {
		strcat(flags, "M");
	}
	if (ch->flags & M_CHAN_L) {
		strcat(flags, "L");
	}
	for (p = connections; p; p = p->next) {
		if (!p->locked && (p->type == CT_HOST)) {
			sprintf(buffer, "/\377\200MODE %d -sptiml+%s\n", ch->chan, flags);
			linklog(p, L_SENT, "%s", &buffer[3]);
			appendstring(p, buffer);
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_opermsg(char *toname, char *hostname, char *fromname, int channel)
{
	char buffer[2048];
	struct connection *p;

	for (p = connections; p; p = p->next) {
		if (!p->locked) {
			if (p->type == CT_HOST) {
				sprintf(buffer, "/\377\200OPER %s %d %s\n", fromname, channel, toname);
				linklog(p, L_SENT, "%s", &buffer[3]);
				appendstring(p, buffer);
				p->locked = 1;
			} else {
				if (p->type == CT_USER) {
					if (!p->via && (channel != -1)) {
						if ((!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)) && (p->channel == channel)) {
							appendc(p, 0, 0);
							appendc(p, 3, 0);
							sprintf(buffer, "*** (%s) %s made you a channel operator for channel %d\n",
								ts2(currtime), fromname, channel);
							appendstring(p, buffer);
							appendprompt(p, 0);
							p->locked = 1;
						} else {
							if (p->verbose) {
								appendc(p, 0, 0);
								appendc(p, 3, 0);
								sprintf(buffer, "*** (%s) %s@%s is now a channel operator for channel %d\n",
									ts2(currtime), toname, hostname, channel);
								appendstring(p, buffer);
								appendprompt(p, 0);
								p->locked = 1;
							}
						}
					}
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_persmsg(char *fromname, char *hostname, int channel, char *text, time_t time)
{
	char buffer[2048];
	struct connection *p;
	struct channel *ch;
	char chan[128];
	int mychannel;
	struct clist *cl;

	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == channel)
			break;
	}
	sprintf(chan, "channel %d", channel);
	if (ch) {
		if (ch->flags & M_CHAN_S)
			sprintf(chan, "secret channel %d", channel);
		if (ch->flags & M_CHAN_I)
			sprintf(chan, "invisible channel %d", channel);
	}
	for (p = connections; p; p = p->next) {
		if (p->type == CT_HOST) {
			if (!p->locked) {
				if (p->amprnet)
					sprintf(buffer, "/\377\200USER %s %s %ld %d %d %s\n", fromname, hostname, (long)time, channel, channel, text);
				else
					sprintf(buffer, "/\377\200UDAT %s %s %s\n", fromname, hostname, text);
				linklog(p, L_SENT, "%s", &buffer[3]);
				appendstring(p, buffer);
				p->locked = 1;
			}
		} else {
			mychannel = -1;
			for (cl = p->chan_list; cl; cl = cl->next) {
				if (cl->channel == channel) {
					mychannel = channel;
					break;
				}
			}
			if ((p->type == CT_USER) && !p->locked && !p->via &&
			    (p->verbose || (mychannel == channel))) {
				appendc(p, 0, 0);
				appendc(p, 3, 0);
				if ((*text != '\0') && (strcmp(text, "@"))) {
					sprintf(buffer, "*** (%s) %s@%s on %s set personal text:\n    %s\n", ts2(currtime), fromname, hostname, chan, text);
				} else {
					sprintf(buffer, "*** (%s) %s@%s on %s removed personal text.\n", ts2(currtime), fromname, hostname, chan);
				}
				appendstring(p, buffer);
				appendprompt(p, 0);
			}
			p->locked = 1;
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_topic(char *fromname, char *hostname, time_t time, int channel, char *text)
{
	char buffer[2048];
	struct connection *p;
	struct channel *ch;
	char chan[128];
	char *chr = text;
	int mychannel, flags = 0;
	struct clist *cl;

	for (ch = channels; ch; ch = ch->next)
		if (ch->chan == channel)
			break;
	if (ch) {
		sprintf(chan, "channel %d", channel);
		flags = ch->flags;
		if (flags & M_CHAN_S)
			sprintf(chan, "secret channel %d", channel);
		if (flags & M_CHAN_I)
			sprintf(chan, "invisible channel %d", channel);
		if (ch->time < time) {

			while (*chr != '\0') {	/* Cut out bells */
				if (*chr == '\7')
					*chr = ' ';
				chr++;
			}

			strncpy(ch->topic, text, TOPICSIZE);
			strncpy(ch->tsetby, fromname, NAMESIZE);
			ch->time = time;
			for (p = connections; p; p = p->next) {
				if (p->type == CT_HOST) {
					if (!p->locked) {
						sprintf(buffer, "/\377\200TOPI %s %s %ld %d %s\n", fromname, hostname, (long)time, channel, text);
						linklog(p, L_SENT, "%s", &buffer[3]);
						appendstring(p, buffer);
						p->locked = 1;
					}
				} else {
					mychannel = -1;
					for (cl = p->chan_list; cl; cl = cl->next) {
						if (cl->channel == channel) {
							mychannel = channel;
							break;
						}
					}
					if ((p->type == CT_USER) && !p->locked && !p->via &&
					    ((!(ch->flags & M_CHAN_I) && p->verbose) || (mychannel == channel))) {
						appendc(p, 0, 0);
						appendc(p, 3, 0);
						if (*text != '\0')
							sprintf(buffer, "*** (%s) %s@%s on %s set channel topic:\n    %s\n", ts2(currtime), fromname, hostname, chan, text);
						else
							sprintf(buffer, "*** (%s) %s@%s on %s removed channel topic.\n", ts2(currtime), fromname, hostname, chan);
						appendstring(p, buffer);
						appendprompt(p, 0);
					}
					p->locked = 1;
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

int get_hash(char *name)
{
	int i, hash;

	hash = 0;
	for (i = 0; i < strlen(name); i++)
		hash = ((hash << 1) + name[i]) & 0xff;
	return hash;
}


/*---------------------------------------------------------------------------*/

void log_user_change(char *name, char *host, int oldchannel, int newchannel)
{
#if WANT_LOG
	char fname[128];
	sprintf(fname, "%s.%02x", userfile, get_hash(name));
	if (oldchannel < 0 || newchannel < 0) {
		FILE *f;
		long found;
		char *p, *q;
		char xname[512];

		f = fopen(fname, "r+");
		if (f == NULL) {
			f = fopen(fname, "w+");
			if (f == NULL)
				return;
		}
		q = xname;
		for (p = name; isalnum(*p); p++)
			*q++ = *p;
		*q++ = '@';
		*q = '\0';
		strcat(xname, host);

		/* zuerst die passende zeile suchen */
		found = -1l;
		while (!feof(f)) {
			char buffer[512];
			found = ftell(f);
			fgets(buffer, 512, f);
			if (!strncmp(buffer, xname, strlen(xname))	/*  && 
									   ((newchannel < 0 && strstr(buffer," +")) ||
									   (oldchannel < 0 && strstr(buffer," -")))       */ )
				break;
		}

		rewind(f);
		if (found >= 0l) {
			fseek(f, found, SEEK_SET);
		} else {
			fclose(f);
			f = fopen(userfile, "a");
		}

		if (f != NULL) {
			struct tm *tm;

			tm = localtime(&currtime);
			fprintf(f, "%-20s %2d.%02d.%04d %2d:%02d %c\n",
				xname, tm->tm_mday, tm->tm_mon + 1, tm->tm_year + 1900, tm->tm_hour, tm->tm_min,
				(oldchannel < 0) ? '+' : '-');
		}
		fclose(f);
	}
#endif
}

/*---------------------------------------------------------------------------*/

void send_user_change_msg(char *name, char *host, int oldchannel, int newchannel, 
	char *pers, time_t time, int observer)
{
	char buffer[2048];
	struct connection *p;
	char oldchan[128], newchan[128];
	struct channel *ch;
	int oldflags = 0, newflags = 0;
	int mychannel;
	struct clist *cl;

	sprintf(newchan, "channel %i", newchannel);
	sprintf(oldchan, "channel %i", oldchannel);
	for (ch = channels; ch; ch = ch->next)
		if (ch->chan == newchannel)
			break;
	if (ch) {
		newflags = ch->flags;
		if (newflags & M_CHAN_S) {
			sprintf(newchan, "secret channel %d", newchannel);
		}
		if (newflags & M_CHAN_I) {
			sprintf(newchan, "invisible channel %d", newchannel);
		}
	}
	for (ch = channels; ch; ch = ch->next)
		if (ch->chan == oldchannel)
			break;
	if (ch) {
		oldflags = ch->flags;
		if (oldflags & M_CHAN_S) {
			sprintf(oldchan, "secret channel %d", oldchannel);
		}
		if (oldflags & M_CHAN_I) {
			sprintf(oldchan, "invisible channel %d", oldchannel);
		}
	}
	log_user_change(name, host, oldchannel, newchannel);
	for (p = connections; p; p = p->next) {
		mychannel = -1;
		if (p->type == CT_USER && !p->via && !p->locked) {
			for (cl = p->chan_list; cl; cl = cl->next) {
				if (cl->channel == oldchannel) {
					mychannel = oldchannel;
					break;
				}
			}
			if ((newchannel == oldchannel) && (newchannel != -1)) {
				if ((p->verbose && !(newflags & M_CHAN_I)) || (mychannel == newchannel)) {
					appendc(p, 0, 0);
					appendc(p, 3, 0);
					if ((*pers != '\0') && (strcmp(pers, "@")))
						sprintf(buffer, "*** (%s) %s@%s on %s set personal text:\n    %s\n", ts2(currtime), name, host, newchan, pers);
					else
						sprintf(buffer, "*** (%s) %s@%s on %s removed personal text.\n", ts2(currtime), name, host, newchan);
					appendstring(p, buffer);
					appendprompt(p, 0);
					p->locked = 1;
				}
			} else {
				if (oldchannel >= 0) {
					if ((!(oldflags & M_CHAN_I) && p->verbose) || (mychannel == oldchannel)) {
						appendc(p, 0, 0);
						appendc(p, 3, 0);
						if (pers && (*pers != '\0') && strlen(pers) < 2) {
							sprintf(buffer, "*** (%s) %s@%s left %s.\n", ts2(currtime), name, host, oldchan);
						} else {
							sprintf(buffer, "*** (%s) %s@%s left %s (%s).\n", ts2(currtime), name, host, oldchan, pers);
						}
						log(L_NUSER, "%s@%s left %i", name, host, oldchannel);
						appendstring(p, buffer);
						appendprompt(p, 0);
						p->locked = 1;
					}
				}
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cl->channel == newchannel) {
						break;
					}
				}
				if (newchannel >= 0) {
					sprintf(buffer, " %s ", name);
					if ((cl) || (!(newflags & M_CHAN_I) && (p->verbose || strstr(p->notify, buffer)))) {
						appendc(p, 0, 0);
						appendc(p, 3, 0);
						sprintf(buffer, "*** (%s) %s@%s joined %s\n", ts2(currtime), name, host, newchan);
						appendstring(p, buffer);
						log(L_NUSER, "%s@%s joined %s", name, host, newchan);
						appendprompt(p, 0);
						p->locked = 1;
					}
				}
			}
		}
		if (p->type == CT_HOST && !p->locked) {
			if (time == currtime)
				time++;
			if (p->amprnet) {
				sprintf(buffer, "/\377\200%s %s %s %ld %d %d %s\n", (observer) ? "OBSV" : "USER", name, host, (long)time, oldchannel, newchannel, pers);
			} else {
				sprintf(buffer, "/\377\200%s %s %s %ld %d %d\n", (observer) ? "OBSV" : "USER", name, host, (long)time, oldchannel, newchannel);
				if (newchannel != -1) {
					if ((oldchannel != -1) || ((pers) && (*pers) && (*pers != '@'))) {
						if (!observer) {
							linklog(p, L_SENT, "%s", &buffer[3]);
							appendstring(p, buffer);
							sprintf(buffer, "/\377\200UDAT %s %s %s\n", name, host, pers);
						}
					}
				}
			}
			linklog(p, L_SENT, "%s", &buffer[3]);
			appendstring(p, buffer);
			p->locked = 1;
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_msg_to_local_user(char *fromname, char *toname, char *text)
{

	char buffer[2048];
	struct connection *p;
	char buf[(NAMESIZE * 2) + 2];

	for (p = connections; p; p = p->next)
		if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)) && !p->locked)
			if (!p->via) {
				if (!p->locked) {
					appendc(p, 0, 0);
					appendc(p, 3, 0);
					if (strcmp(fromname, "conversd")) {
						sprintf(buffer, "<*%s*>:", fromname);
						appendstring(p, formatline(buffer, text, p->width));
						if (p->away[0] != '\0' && text[0] != '*') {
							getTXname(p, buf);
							sprintf(buffer, "*** (%s) %s is away: %s", ts2(currtime), buf, p->away);
							send_msg_to_local_user("conversd", fromname, buffer);
						}
					} else {
						appendstring(p, text);
						appendstring(p, "\n");
					}
					appendprompt(p, 0);
					p->locked = 1;
				}
			}
}

/*---------------------------------------------------------------------------*/

void send_msg_to_user(char *fromname, char *toname, char *text)
{

	char buffer[2048];
	struct connection *p;
	char buf[(NAMESIZE * 2) + 2];

	for (p = connections; p; p = p->next)
		if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)) && !p->locked) {
			if (p->via) {
				if (!p->via->locked) {
					sprintf(buffer, "/\377\200UMSG %s %s %s\n", fromname, toname, text);
					linklog(p, L_SENT, "%s", &buffer[3]);
					appendstring(p->via, buffer);
					p->via->locked = 1;
					p->locked = 1;
				}
			} else {
				if (!p->locked) {
					appendc(p, 0, 0);
					appendc(p, 3, 0);
					if (strcmp(fromname, "conversd")) {
						sprintf(buffer, "<*%s*>:", fromname);
						appendstring(p, formatline(buffer, text, p->width));
						if (p->away[0] != '\0' && text[0] != '*') {
							getTXname(p, buf);
							sprintf(buffer, "*** (%s) %s is away: %s", ts2(currtime), buf, p->away);
							send_msg_to_local_user("conversd", fromname, buffer);
						}
					} else {
						appendstring(p, text);
						appendstring(p, "\n");
					}
					appendprompt(p, 0);
					p->locked = 1;
				}
			}
		}
}

/*---------------------------------------------------------------------------*/

void send_msg_to_channel(char *fromname, int channel, char *text)
{
	char buffer[2048];
	struct connection *p;
	struct channel *ch;
	char addchan[16];
	struct clist *cl;
	int printit;

	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == channel)
			break;
	}
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER && (!strcasecmp(p->name, fromname) || !strcasecmp(p->nickname, fromname))) {
			p->mtime = currtime;
		}
	}
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER) {
			printit = 0;
			if (p->channel == channel) {
				printit = 1;
				addchan[0] = '\0';
			} else {
				for (cl = p->chan_list; cl; cl = cl->next) {
					if (cl->channel == channel) {
						printit = 1;
						sprintf(addchan, "%d:", channel);
					}
				}
			}
			if (printit) {
				if (p->via) {
					if (!p->via->locked && !(ch->flags & M_CHAN_L)) {
						sprintf(buffer, "/\377\200CMSG %s %d %s\n", fromname, channel, text);
						linklog(p, L_SENT, "%s", &buffer[3]);
						appendstring(p->via, buffer);
						p->via->locked = 1;
					}
				} else {
					if (!p->locked) {
						appendc(p, 0, 0);
						appendc(p, 3, 0);
						if (strcmp(fromname, "conversd")) {
							sprintf(buffer, "<%s%s>:", addchan, fromname);
							appendstring(p, formatline(buffer, text, p->width));
						} else {
							appendstring(p, text);
							appendstring(p, "\n");
						}
						appendprompt(p, 0);
						p->locked = 1;
					}
				}
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

void send_invite_msg(char *fromname, char *toname, int channel)
{

	static char invitetext[] = "\n\007\007*** (%s) Message from %s...\nPlease join %s channel %d.\n\007\007\n";

	static char responsetext[] = "*** (%s) %s Invitation sent to %s @ %s.";

	char buffer[2048];
	char convtype2[2048];
	int fdtty;
	int fdut;
	struct connection *p;
	struct stat stbuf;
	struct utmp utmpbuf;

	strncpy(convtype2, convtype, 2048);
	convtype2[strlen(convtype2) - 1] = '\0';
	for (p = connections; p; p = p->next) {
		if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname))) {
			if (p->channel == channel) {
				clear_locks();
				sprintf(buffer, "*** (%s) User %s is already on this channel.", ts2(currtime), toname);
				send_msg_to_user("conversd", fromname, buffer);
				return;
			}
			if (!p->via && !p->locked) {
				sprintf(buffer, invitetext, ts2(currtime), fromname, convtype2, channel);
				p->invitation_channel = channel;
				appendstring(p, buffer);
				clear_locks();
				sprintf(buffer, responsetext, ts2(currtime), convtype2, toname, myhostname);
				send_msg_to_user("conversd", fromname, buffer);
				return;
			}
			if (p->via && !p->via->locked) {
				sprintf(buffer, "/\377\200INVI %s %s %d\n", fromname, toname, channel);
				linklog(p, L_SENT, "%s", &buffer[3]);
				appendstring(p->via, buffer);
				return;
			}
		}
	}

	if ((fdut = open("/etc/utmp", O_RDONLY, 0644)) >= 0) {
		int found = 0;
		while (read(fdut, (char *) &utmpbuf, sizeof(utmpbuf)) == sizeof(utmpbuf)) {
			if (!strncmp(utmpbuf.ut_name, toname, sizeof(utmpbuf.ut_name))) {
				found++;
				strcpy(buffer, "/dev/");
				strncat(buffer, utmpbuf.ut_line, sizeof(utmpbuf.ut_line));
				if (stat(buffer, &stbuf))
					continue;
				if (!(stbuf.st_mode & 2))
					continue;
				if ((fdtty = open(buffer, O_WRONLY | O_NOCTTY, 0644)) < 0)
					continue;
				sprintf(buffer, invitetext, ts2(currtime), fromname, convtype2, channel);
				if (!fork()) {
					write(fdtty, buffer, strlen(buffer));
					exit(0);
				}
				close(fdtty);
			}
		}
		if (found) {
			clear_locks();
			sprintf(buffer, responsetext, ts2(currtime), convtype2, toname, myhostname);
			send_msg_to_user("conversd", fromname, buffer);
			close(fdut);
			return;
		}
		close(fdut);
	}
	for (p = connections; p; p = p->next) {
		if (p->type == CT_HOST && !p->locked) {
			sprintf(buffer, "/\377\200INVI %s %s %d\n", fromname, toname, channel);
			linklog(p, L_SENT, "%s", &buffer[3]);
			appendstring(p, buffer);
		}
	}
}


/*---------------------------------------------------------------------------*/

void update_destinations(struct permlink *p, char *name, long rtt, char *rev)
{
	struct permlink *l;
	struct destination *d, *d1;
	struct connection *cp;
	int pl;
	char buffer[2048];
	int force_send = 0;

	if(!permarray || !rev)
		return; 

	for (d = destinations; d; d = d->next)
		if (!strcasecmp(d->name, name))
			break;

//	hops++;

	if (!d) {
		d = (struct destination *)hmalloc(sizeof(struct destination));
		if (d) {
			d->last_sent_rtt = 0;
			d->link = p;
			d->rtt = 0;
			strncpy(d->name, name, HOSTNAMESIZE);
			d->name[sizeof(d->name)-1] = 0;
			strncpy(d->rev, rev, NAMESIZE);
			d->rev[sizeof(d->rev)-1] = 0;
			if (!destinations || (!destinations->name) || (strcasecmp(destinations->name, name) > 0)) {
				d->next = destinations;
				destinations = d;
			} else {
				d1 = destinations;
				while (d1->next) {
					if (strcasecmp(d1->next->name, name) > 0) {
						d->next = d1->next;
						d1->next = d;
						break;
					}
					d1 = d1->next;
				}
				if (!d1->next) {
					d->next = d1->next;
					d1->next = d;
				}
			}
			force_send = 1;
		}
	} else {
		if (d->link != NULLPERMLINK && d->link != p && d->rtt && (rtt >= d->rtt)) {
			log(L_R_ERR, "CMD ignored from %s: /..DEST %s %ld %s",
			      p->name, name, rtt, rev);
			return;
		}
		if (strcmp(rev, "-") && (rtt || (!rtt && *rev))) {
			/* DL9SAU:
		         * I. passive-learned new destinations hack.
                         * all known convers derivates will override rev
                         * when a new rev will occour, so we could send this
                         * information until we have received a real //DEST
                         * but do not overwrite if we have already a more
                         * useful information..
                         * II. //DEST 0 is usualy sent without rev-info - in
                         * this case, keep the old revision information.
                         */
		strncpy(d->rev, rev, NAMESIZE);
		d->rev[sizeof(d->rev)-1] = 0;
	}
	if (d->link == NULLPERMLINK || p == NULLPERMLINK || d->link != p)
		force_send = 1;
		d->link = p;
	}

	if (!d->rtt && rtt) {
		for (cp = connections; cp; cp = cp->next) {
			if (!strcasecmp(cp->name, name))
				break;
		}
		if (cp == NULLCONNECTION)
			notify_destinations(d, name, rev, 1);
			force_send = 1;
	}
	d->rtt = rtt;

	if (abs(rtt - d->last_sent_rtt) > (d->last_sent_rtt / 8))
		force_send = 1;

	if (rtt && d->last_sent_rtt && p != NULLPERMLINK && d->link == p && (p->connection) && (p->connection->type == CT_HOST) &&
		!strcasecmp(d->name, name) && !strcasecmp(d->link->name, name)) {
		for (d1 = destinations; d1; d1 = d1->next) {
			if (!d1->rtt || d1->link == NULLPERMLINK ||
					d1 == d || d1->link == p ||
					!*d1->name || !strcasecmp(d1->name, name) || !strcasecmp(d1->link->name, name))
				continue;
			if (abs((d1->rtt + d->rtt) - (d1->rtt + d->last_sent_rtt)) > ((d1->last_sent_rtt + d->last_sent_rtt) / 3)) {
				sprintf(buffer, "/\377\200DEST %s %ld %s\n", d1->name, d1->rtt + (p->txtime + p->rxtime) / 3L, d1->rev);
				linklog(p->connection, L_SENT, "%s", &buffer[3]);
				log(L_DEBUG, "debug: updated %s to %s", &buffer[3], p->name);
				appendstring(p->connection, buffer);
				d1->last_sent_rtt = d1->rtt;
			}
		}
	}
	if (!force_send) {
		return;
	}
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		if (!(l = permarray[pl]))
			continue;
		if (!strcasecmp(d->name, l->name) || (d->link && !strcasecmp(d->link->name, l->name)))
			continue;
		if (l->connection && (l->connection->type == CT_HOST) && (d->link == NULLPERMLINK || d->link != p || strcasecmp(d->link->name, l->name))) {
			if (rtt) {
				sprintf(buffer, "/\377\200DEST %s %ld %s\n", name, rtt + (l->txtime + l->rxtime) / 2L, d->rev);
			} else {
				sprintf(buffer, "/\377\200DEST %s 0\n", name);
			}
			linklog(l->connection, L_SENT, "%s", &buffer[3]);
			log(L_DEBUG, "debug: sent %s to %s", &buffer[3], l->name);
			appendstring(l->connection, buffer);
		}
	}
	d->last_sent_rtt = rtt;
	if (!rtt)
		d->link = NULLPERMLINK;

}

/*---------------------------------------------------------------------------*/

struct permlink *update_permlinks(char *name, struct connection *cp, int isperm)
{
	struct destination *d;
	struct permlink *p = NULL;
	int pl = -1;
	
	if (!permarray)
		permarray = hcalloc(NR_PERMLINKS, sizeof(struct permlink *));
		
	if (cp != NULLCONNECTION)
		notify_destinations(NULLDESTINATION, name, cp->rev, 1);
		
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		p = permarray[pl];
		if (p && !strcasecmp(p->name, name)) {
			for (d = destinations; d; d = d->next) {
				if (!isperm && (d->link == p) && cp == NULLCONNECTION)
					notify_destinations(d, d->name, d->rev, 0);
				if (d->rtt && (d->link == p))
					update_destinations(p, d->name, 0, "");
			}
			if (cp) {
				p->connection = cp;
				p->statetime = currtime;
				p->tries = 0;
				p->waittime = 9;
				p->rxtime = 0;
				p->txtime = 0;
				p->testwaittime = currtime;
				p->testnexttime = currtime + 60;
				p->retrytime = currtime + p->waittime;
				if (isperm)
					p->permanent = isperm;
			} else {
				if (!isperm && cp == NULLCONNECTION) {
					log(L_LINK, "Disconnected from %s", p->name);
					notify_destinations(NULLDESTINATION, p->name, p->connection->rev, 0);
				}
				if (!p->permanent) {
					hfree(p);
					permarray[pl] = NULLPERMLINK;
				}
			}
			return p;
		}
	}
	
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		if (!permarray[pl])
			break;
	}
	
	if ((pl > -1) && (pl < NR_PERMLINKS)) {
		p = (struct permlink *)hcalloc(1, sizeof(struct permlink));
		strncpy(p->name, name, HOSTNAMESIZE);
		p->host = NULL;
		p->permanent = 0;
		p->connection = cp;
		p->statetime = currtime;
		p->tries = 0;
		p->waittime = 9;
		p->rxtime = 0;
		p->txtime = 0;
		p->testnexttime = currtime + 60;
		p->testwaittime = currtime;
		p->retrytime = currtime + p->waittime;
		p->permanent = isperm;
		permarray[pl] = p;
	} else {
		bye_command(cp);	/* no space left in array */
		return NULLPERMLINK;
	}
	
	return p;
}

/*---------------------------------------------------------------------------*/

void connect_permlinks(void)
{

#define MAX_WAITTIME   (60*60)

	char buffer[2048];
	int addrlen;
	int fd;
	int pl;
	struct connection *cp;
	struct permlink *p, *p2;
	struct sockaddr_in addr;
	struct destination *d;
	struct hostent *hp;
	
	/* Update our existing connections */
	if (!permarray)
		return;
		
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		p = permarray[pl];
		if (p && p->connection) {
			if (p->testnexttime < currtime) {
				if ((p->testwaittime + 7300) < currtime) {
					p->rxtime = 0;
					p->txtime = 0;
					for (d = destinations; d; d = d->next) {
						if (d->link == p) {
							update_destinations(p, d->name, 0, "");
						}
					}
				}
				linklog(p->connection, L_SENT, "PING\n");
				appendstring(p->connection, "/\377\200PING\n");
				p->testwaittime = currtime;
				p->testnexttime = currtime + 7300;
			}
		}
	}

	/* Now attempt to connect our down connections */
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		p = permarray[pl];
		if (p && !p->connection) {
			if (p->retrytime < currtime) {
				/* First see if there are any primary links. If there are and
				 * they are connected, don't connect this link.
				 */
				for (p2 = p->primary; p2; p2 = p2->primary) {
					if (p2->connection)
						break;
				}
				if (p2 && p2->connection) {
					p->retrytime = currtime;
					p->waittime = 9;
					continue;
				}
				/* Ok, no primary connections, let's try and connect */

				p->tries++;
				if (p->waittime == 9)
					p->waittime = 60;	/* *2 = 10 minutes */
				p->waittime <<= 1;	/* 10,20,40,80,160,180,180,... min */
				if (p->waittime > MAX_WAITTIME)
					p->waittime = MAX_WAITTIME;
				p->retrytime = currtime + p->waittime;
				
				hp = gethostbyname(p->host);
				if (hp == NULL) {
					log(L_ERR, "connect_permlinks(): Unknown host: %s", p->host);
					continue;
				}
				
				addrlen = sizeof(struct sockaddr_in);
				memset(&addr, 0, sizeof(struct sockaddr_in));
				addr.sin_family = AF_INET;
				addr.sin_port = htons(p->port);
				memcpy(&addr.sin_addr, hp->h_addr_list[0], hp->h_length);
				
				if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
					log(L_ERR, "connect_permlinks(): socket() failed: %s", strerror(errno));
					continue;
				}
				
				if (fcntl(fd, F_SETFL, O_NONBLOCK) == -1) {
					log(L_ERR, "connect_permlinks(): fcntl(O_NONBLOCK) failed: %s", strerror(errno));
					close(fd);
					continue;
				}
				
				if (connect(fd, (struct sockaddr *)&addr, addrlen)) {
					if (errno != EINPROGRESS) {
						log(L_ERR, "connect_permlinks(): connect to %s failed: %s", p->name, strerror(errno));
						close(fd);
						continue;
					}
				}
				cp = alloc_connection(fd);
				if (!cp)
					log(L_ERR, "connect_permlinks(): cp == 0 on %s connect", p->name);
				p->connection = cp;
				
				log(L_LINK, "Connecting to %s ...", p->name);
				
				/* We have made a connection. Now see if there are any backup
				 * links, and if so, close them down.
				 
				for (p2 = p->backup; p2; p2 = p2->backup)
					if (p2->connection) {
						bye_command2(p2->connection, "backup link closed");
						free_closed_connections();
						sleep(10); EVIL, should not sleep, just send the HOST much later 
					}
*/				
				sprintf(buffer, "/\377\200HOST %s %s %s\n", myhostname, SOFT_REL8, myfeatures);
				linklog(cp, L_SENT, "%s", &buffer[3]);
				appendstring(cp, buffer);
				p->testnexttime = currtime + 5;
			}
		}
	}
}

/*---------------------------------------------------------------------------*/

static void process_input(struct connection *cp)
{
	char *arg;
	char buffer[2048];
	int arglen;
	struct cmdtable *cmdp;
	struct channel *ch;
	char buf[(NAMESIZE * 2) + 2];

	clear_locks();
	cp->locked = 1;

	if ((cp->type == CT_USER) && (strncmp(cp->ibuf, "/\377\200", 3)))
		convert(cp->charset_in, ISO, cp->ibuf);

	if (*cp->ibuf == '/') {
		if (cp->ibuf[1] == '\377' && cp->ibuf[2] == '\200')
			linklog(cp, L_RECV, "%s\n", &cp->ibuf[3]);
		arglen = strlen(arg = getarg(cp->ibuf + 1, 0));
		for (cmdp = cmdtable; cmdp->name; cmdp++)
			if (!strncmp(cmdp->name, arg, arglen)) {
				if (cp->observer && !(cmdp->states & CM_OBSERVER)) {
					appendstring(cp, OBSERVERSTR);
					appendstring(cp, OBSERVERSTR2);
					appendprompt(cp, 0);
					return;
				}
				if (!cp->observer || cmdp->states & CM_OBSERVER)
					if (cmdp->states & (1 << cp->type))
						(*cmdp->fnc) (cp);
				return;
			}
		/* didn't find match in table - see if it is a host command */
		if (cp->ibuf[1] == '\377' && cp->ibuf[2] == '\200') {
			*(--getargp) = ' ';
			h_unknown_command(cp);	/* yep, pass it on */
			return;
		}
		if (cp->type == CT_USER) {
			if (!ecmd_exists(arg)) {
				sprintf(buffer, "*** Unknown command '/%s'. Type /HELP for help.\n", arg);
				appendstring(cp, buffer);
			} else {
				char *params;
				params = getarg(0, 1);
				sprintf(buffer, "/\377\200ECMD %s %s %s\n", cp->name, arg, params);
				strcpy(cp->ibuf, buffer);
				h_unknown_command(cp);
			}
			appendprompt(cp, 0);
		}
		return;
	}
	if (cp->observer)
		appendstring(cp, OBSERVERSTR);
	else if (cp->type == CT_USER) {
		getTXname(cp, buf);
		if (*cp->away)
			appendstring(cp, "*** You are away, aren't you ? :-)\n");
		if (cp->expected) {
			getarg(cp->ibuf, 1);	/* just fill that static variable */
			operator_command(cp);
			cp->expected = 0;
		} else {
			if (cp->query[0] == '\0') {
				for (ch = channels; ch; ch = ch->next) {
					if (ch->chan == cp->channel)
						break;
				}
				if (cp->operator || cp->channelop || !(ch->flags & M_CHAN_M))
					send_msg_to_channel(buf, cp->channel, cp->ibuf);
				else
					appendstring(cp, "*** This is a moderated channel. Only channel operators may write.\n");
			} else
				send_msg_to_user(buf, cp->query, cp->ibuf);
		}
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

static int ours(const char *buf, const char *search)
{
	char buf2[82], search2[82];

	if (strchr(buf, '*'))
		return ((search == NULLCHAR) ? 1 : 0);
	if (search == NULLCHAR)
		return 1;
	strcpy(buf2, buf);
	strcpy(&search2[1], search);
	search2[0] = ':';
	strlwr(buf2);
	strlwr(search2);
	return (strstr(buf2, search2) != NULL);
}

/*---------------------------------------------------------------------------*/

int dohelper(char *name, const char *search)
{
	char buf[82], buf2[100];
	int count;
	FILE *fp;
	int foundit = 0;
	int didheader = 0;

	clear_locks();
	if ((fp = fopen(helpfile, "r")) != (FILE *) 0) {
		if (search == NULLCHAR)
			didheader = 1;
		count = 0;
		do {
			if (fgets(buf, 80, fp)) {
				if (*buf == '#')
					continue;	/* skip comments */
				if (*buf != ':') {
					if (count) {
						count--;
						rip(buf);
						if (!didheader) {
							send_msg_to_user("conversd", name, "*** Usage:");
							clear_locks();
							didheader = 1;
						}
						sprintf(buf2, "%s%s", (*buf != '/') ? "  " : "", buf);
						send_msg_to_user("conversd", name, buf2);
						clear_locks();
					}
				} else {	/* start of field or continuation */
					count = (ours(buf, search)) ? 32000 : 0;
					if (search == NULLCHAR && count)
						count = 1;
					if (buf[1] == ':' && count == 1)
						count--;
					if (count)
						foundit = 1;
				}
			}
		} while (!feof(fp));
		fclose(fp);
	}
	return (foundit);
}

/*---------------------------------------------------------------------------*/

void buildfnames(char *confdir)
{
	sprintf(conffile, "%s/convers.conf", confdir);
	sprintf(motdfile, "%s/convers.motd", confdir);
	sprintf(restrictfile, "%s/convers.restrict", confdir);
	sprintf(accessfile, "%s/convers.access", confdir);
	sprintf(noaccessfile, "%s/convers.noaccess", confdir);
	sprintf(helpfile, "%s/convers.help", confdir);
	sprintf(observerfile, "%s/sconvers.observer", confdir);
	sprintf(issuefile, "%s/convers.issue", confdir);
	sprintf(userfile, "%s/log/convers.log", DATA_DIR);
	sprintf(pidfile, "/var/run/conversd.pid");
#if 0
	printf("conffile: '%s'\n", conffile);
	printf("motdfile: '%s'\n", motdfile);
	printf("restrictfile: '%s'\n", restrictfile);
	printf("accessfile: '%s'\n", accessfile);
	printf("noaccessfile: '%s'\n", noaccessfile);
	printf("helpfile: '%s'\n", helpfile);
	printf("observerfile: '%s'\n", observerfile);
	printf("issuefile: '%s'\n", issuefile);
	printf("userfile: '%s'\n", userfile);
	fflush(stdout);
#endif
	return;
}


/*---------------------------------------------------------------------------*/

void update_stats(time_t now)
{
	int k;
	
	if (now >= nexthourlystats) {
		for (k = 59; k; k--) {
			hourly[k].tx = hourly[k - 1].tx;
			hourly[k].rx = hourly[k - 1].rx;
			hourly[k].start = hourly[k - 1].start;
		}
		hourly[0].start = nexthourlystats;
		nexthourlystats += ONE_HOUR;
		hoursonline++;
		hourly[0].rx = hourly[0].tx = 0;
	}
	if (now >= nextdailystats) {
		for (k = 59; k; k--) {
			daily[k].tx = daily[k - 1].tx;
			daily[k].rx = daily[k - 1].rx;
			daily[k].start = daily[k - 1].start;
		}
		daily[0].start = nextdailystats;
		nextdailystats += ONE_DAY;
		daysonline++;
		daily[0].rx = daily[0].tx = 0;
	}
}

/*---------------------------------------------------------------------------*/

void init_stats(void)
{
	struct tm *t;

	starttime = time((time_t *) 0);
	t = localtime(&starttime);
	t->tm_min = t->tm_sec = 0;
	hourly[0].start = mktime(t);
	nexthourlystats = hourly[0].start + ONE_HOUR;
	if (t->tm_hour)
		t->tm_hour = 0;
	daily[0].start = mktime(t);
	nextdailystats = daily[0].start + ONE_DAY;
	daily[0].rx = daily[0].tx = hourly[0].rx = hourly[0].tx = 0;
}

/*---------------------------------------------------------------------------*/

void mysighup(int sig)
{
	struct connection *p;
	
	/* bring down all the links, users and hosts */
	for (p = connections; p; p = p->next)
		bye_command2(p, "link shutdown");
	/* now actually do the disconnects */
	free_closed_connections();
	unlink(pidfile);
	if (sig == SIGHUP)
		log(L_INFO, "SIGHUP received - restarting process");
	signal(SIGHUP, mysighup);
}

/* -----------------------------------------------------------------------
 *	Write PID to file (to aid external scripts kick our butt)
 */
 
void write_pid(void)
{
	int f;
	char s[10];
	
	sprintf(s, "%d\n", getpid());
	s[8] = '\0';
	
	if ((f = creat(pidfile, 0644)) < 0) {
		log(L_CRIT, "Could not open pid file %s for writing: %s", pidfile, strerror(errno));
		exit(1);
	}
	write(f, s, strlen(s));
	close(f);
}

/* ---------------------------------------------------------------------
 *	Drop to a non-priviledged UID if we happen to be root.
 */

void drop_privileges(void)
{
	uid_t olduid, uid;
	struct passwd *pwent;
	
	if ((olduid = geteuid()) == 0) {
		uid = atoi(OWNER);
		if (uid < 1) {
			if (!((pwent = getpwnam(OWNER))) || pwent->pw_uid == 0) {
				log(L_ERR, "\"%s\" is not a valid non-privileged UID or user name (being root is not preferred).", OWNER);
				return; 
			}
			uid = pwent->pw_uid;
		}
		if (setreuid(uid, uid))
			log(L_ERR, "Could not set real and effective UID's from %d to %d: %s", olduid, uid, strerror(errno));
	}
}

/* ------------------------------------------------------------------------ */

int main(int argc, char **argv)
{
	int listen_fd;
	fd_set actread;
	fd_set actwrite;
	char *sp;
	char buffer[2048];
	int addrlen;
	int arg;
	int i;
	int size;
	int status;
	struct connection *cp;
	struct mbuf *bp;
	struct sockaddr_in addr;
	struct timeval sel_timeout;
	
	drop_privileges();
	parse_cmdl(argc, argv);

	if (demonize) {
		switch (fork()) {
		case 0:	/* child remains running */
			break;
		case -1:	/* error */
			fprintf(stderr, "conversd: fork failed: %s\n", strerror(errno));
			return 1;
			break;
		default:	/* parent - dies a quiet death */
			return 0;
		}
	}
	
	buildfnames(CONF_DIR);
	read_config();
	open_log();
	
	if (tryonly) {
		log(LOG_INFO, "In try-only mode, exiting.");
		return 0;
	}
	
	time(&boottime);
	umask(000);
	
	close(0);
	for (i = 3; i < FD_SETSIZE; i++)
		close(i);
	
	setsid();
	
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, mysighup);
	if (!getenv("TZ"))
		putenv("TZ=localtime");
	
	sp = ctime(&boottime);
	rip(sp);
	log(L_INFO, "%s started at %s", SOFT_RELEASE, sp);
	
	strcpy(convtype, "conversd");
	
	time(&currtime);

	write_pid();
	init_stats();
	
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(listen_port);
	addrlen = sizeof(struct sockaddr_in);
	
	listen_fd = socket(AF_INET, SOCK_STREAM, 0);
	arg = 1;
	setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (char *)&arg, sizeof(arg));
	if (bind(listen_fd, (struct sockaddr *)&addr, addrlen)) {
		log(L_CRIT, "bind() failed: %s", strerror(errno));
		return 1;
	}
	
	if (listen(listen_fd, SOMAXCONN)) {
		log(L_CRIT, "listen() failed: %s", strerror(errno));
		return 1;
	}
	
	if (maxfd < listen_fd) {
		maxfd = listen_fd;
		FD_SET(listen_fd, &chkread);
	}
		
	charset_init();

	for (;;) {

		free_closed_connections();

		while (waitpid(-1, &status, WNOHANG) > 0);
		connect_permlinks();

		actread = chkread;
		actwrite = chkwrite;
		sel_timeout.tv_sec = 60;
		sel_timeout.tv_usec = 0;
		fflush(NULL);
		if (select(maxfd + 1, &actread, &actwrite, (fd_set *) 0, &sel_timeout) <= 0) {
			FD_ZERO(&actread);
			FD_ZERO(&actwrite);
		}
		if (maxfd < 0)
			sleep(1);

		time(&currtime);

		update_stats(currtime);
		
		if (FD_ISSET(listen_fd, &actread))
			accept_connect_request(listen_fd);
		
		for (cp = connections; cp; cp = cp->next) {
			if (cp->type == CT_CLOSED)
				break;
				
			if (cp->fd >= 0 && FD_ISSET(cp->fd, &actread)) {
				size = read(cp->fd, buffer, sizeof(buffer));
				if (size <= 0) {
					log(L_LINK, "Stinky linky with %s (read failed: %s)", cp->name, strerror(errno));
					bye_command2(cp, "stinky linky");
				} else {
					cp->received += size;
					daily[0].rx += size;
					hourly[0].rx += size;
					for (i = 0; i < size; i++)
						switch (buffer[i]) {
						case '\n':
						case '\r':
							if (cp->icnt) {
								cp->ibuf[cp->icnt] = '\0';
								process_input(cp);
								cp->icnt = 0;
							} else {
								appendprompt(cp, 0);
							}
							break;
						case '\377':
							if (cp->icnt > 2) {	/* no <CR> or <LF> on prev line */
								cp->ibuf[cp->icnt - 1] = '\0';
								process_input(cp);
								cp->icnt = 1;
								cp->ibuf[0] = '/';
							}
							/* and fall through */
						default:
							if (cp->icnt < sizeof(cp->ibuf) - 5)
								cp->ibuf[cp->icnt++] = buffer[i];
							break;
						}
				}
			}
			if (cp->fd >= 0 && FD_ISSET(cp->fd, &actwrite)) {
				bp = cp->obuf;
				size = write(cp->fd, bp->data, strlen(bp->data));
				if (size <= 0) {
					log(L_LINK, "Stinky Linky with %s (write failed: %s)", cp->name, strerror(errno));
					bye_command2(cp, "stinky linky");
				} else {
					cp->xmitted += size;
					daily[0].tx += size;
					hourly[0].tx += size;
					bp->data += size;
					if (!*bp->data) {
						cp->obuf = bp->next;
						hfree(bp);
					}
					if (!cp->obuf)
						FD_CLR(cp->fd, &chkwrite);
				}
			}
		}
	}
	
	return 0;
}
